/*
    BOMB, an open-platform NUON game
	
	Programming by Andreas Binner
	
	Sound module
*/

/*
   This file mostly deals with streaming audio.
   
   On Samsung players it uses the CDN50X library to access the CD. To fill
   the streaming audio buffer, the NISE external feeder mechanism is used.
   Since the CDN50X API is not interrupt callable, we need to handle the
   loading the the game main loop!
   
   On other players (Thomson) the standard, BIOS-based streaming audio feature
   of NISE is used. This works fine on a DVD, but the CD support in the newer
   BIOS version was never really tested, certainly not in combination with
   streaming audio! On a DRC300N player it might be safer to disable streaming
   audio by simply leaving out the audio files on the CD!
   
   NOTE: A slightly extended (non-blocking reads, ioctl's) libcdn50x.a library is used!!!
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <nuon/sprite.h>
#include <nuon/bios.h>
#include <nuon/dma.h>
#include <nuon/video.h>
#include <nuon/joystick.h>
#include <nuon/mediaio.h>
#include <nuon/nise.h>
#include <nuon/termemu.h>
#include <sys/ioctl.h>

#include "cdn50x.h"
#include "def.h"
#include "proto.h"

/* Size of the streaming audio buffer in sectors */
#define SECTORS_IN_BUFFER 512

typedef struct
{
	int start,len;
} SONG;

SONG songs[MAX_SONGS];
int numSongs=0;

unsigned char *audiobuffer;
int deviceHandle;
void *iso;
int audSector;
int audStartSector,audNumSectors;
int isXA;
int song;
unsigned char *streamAdr,*lastReadAdr;
int streamCount=0;
int lastReadLSN,lastReadCount;
int streamState=0;
int use_cdn50x=0;

enum
{
	kIdle=0,
	kStartRead=1,
	kWaitForRead=2
};

/*
   Check for player type. Initialize if necessary the CDN50X
   library and then NISE (order is important!)
*/

int InitSound()
{
	int err=0;

	if (strstr(getenv("SYSTEM"),"Samsung"))
		use_cdn50x=1;
	else
		use_cdn50x=0;
	
#ifdef STREAMING_AUDIO
	if (use_cdn50x)
	{
		err=CDInit();
		kprintf("CDInit err=%d\n",err);
	}
#endif

    /* Setup NISE */

   	AUDIOInit();
	
	return err;
}


#ifdef STREAMING_AUDIO

/*
    Feeder for Samsung players. Since CDN50X calls are not
	interrupt callable, we just set flags for the main loop!
*/

void myFeeder(unsigned char *adr, long bytesNeeded)
{
	streamAdr=adr;
	streamCount=bytesNeeded/2048;
	kprintf("feed %d %d\n",streamCount,(int)(streamAdr-audiobuffer)/2048);
	if (!(CDRequest&0x02))
		kprintf("Underrun\n");
	streamState=kStartRead;
}

void dummyFeeder(unsigned char *adr, long bytesNeeded)
{
}

/*
    Look for all "levelXX.pcm" fiel on the CD and store start and length
	(in sectors!)
*/

int GetSongInfo(int n,int *start, int *len)
{
	char file[256];
	int err;


	if (use_cdn50x)
	{
		void *fh;

		sprintf(file,"level%d.pcm",n+1);
		fh=iso_open(iso,file,0,0,&err);
		kprintf("iso_open '%s' fh=%x err=%d\n",file,iso,err);
		if (!err)
		{
			iso_ioctl(iso,fh,ISOGETSTARTSECTOR,start,&err);
			*start-=150;
			iso_ioctl(iso,fh,ISOGETSECTORLENGTH,len,&err);
			iso_ioctl(iso,fh,0x00012345,&isXA,&err);
	    	kprintf("start=%d num=%d XA=%d\n",*start,*len,isXA);
			iso_close(iso,fh,&err);
			return 0;
		}
		return -2;
	}
	else
	{
		int fh;
	
		sprintf(file,"/iso9660/level%d.pcm",n+1);
		fh=_FileOpen(file,0,0,&err);
		kprintf("_FileOpen '%s' fh=%x err=%d\n",file,iso,err);
		if (!err)
		{
			_FileIoctl(fh,ISOGETSTARTSECTOR,start,&err);
			*start-=150;
			_FileIoctl(fh,ISOGETSECTORLENGTH,len,&err);
	    	kprintf("start=%d num=%d\n",start,len);
			_FileClose(fh,&err);
			return 0;
		}
		return -2;
	}
}

/*
	Initialize streaming audio
*/

int InitStreamingAudio()
{
	int err;
	int unused;

	audiobuffer=malloc(SECTORS_IN_BUFFER*2048);

	if (use_cdn50x)
	{
		iso=iso_init(0,&err);
		kprintf("iso_init iso=%x err=%d\n",iso,err);
		if (!iso)
			return 0;
	}
	else
	{
		deviceHandle=_MediaOpen(MEDIA_DVD, "", unused, &unused);
		if (!deviceHandle)
			return 0;
	}
	numSongs=0;
	for(;;)
	{	
		if (GetSongInfo(numSongs,&songs[numSongs].start,&songs[numSongs].len)<0)
			break;
		numSongs++;
		if (numSongs==MAX_SONGS)
			break;
	}
	
	kprintf("%d songs found\n",numSongs);
	
    AUDIOMixer(0x10000000,0x30000000);
	return numSongs;
}

/*
	Start song numner "n"
*/

int StartSong(int n)
{
	if (!numSongs)
		return -1;
		
	n=n%numSongs;
	
	audStartSector=audSector=songs[n].start;
	audNumSectors=songs[n].len;
	kprintf("Song %d: %d-%d\n",n,audStartSector,audStartSector+audNumSectors-1);
	memset(audiobuffer,0,SECTORS_IN_BUFFER*2048);
	
	if (use_cdn50x)
	{
		streamState=kIdle;
		if (audNumSectors<SECTORS_IN_BUFFER)
		{
			AUDIOInitStreamingAudio(deviceHandle,(int)audiobuffer,audNumSectors);
			AUDIOSetupExternalStreamFeeder(dummyFeeder);
			kprintf(" read 0 %d-%d\n",audNumSectors,audSector,audSector+audNumSectors-1);
			CDReadSectors(audSector,audNumSectors,audiobuffer,isXA);
		}
		else
		{
			AUDIOInitStreamingAudio(deviceHandle,(int)audiobuffer,SECTORS_IN_BUFFER);
			AUDIOSetupExternalStreamFeeder(myFeeder);
			kprintf(" read 0 %d-%d\n",audSector,audSector+SECTORS_IN_BUFFER-1);
			CDReadSectors(audSector,SECTORS_IN_BUFFER,audiobuffer,isXA);
		}
	}
	else
	{
		AUDIOInitStreamingAudio(deviceHandle,(int)audiobuffer,SECTORS_IN_BUFFER);
		AUDIOSetupStreamingAudio(audStartSector,audNumSectors);
	}
	audSector+=SECTORS_IN_BUFFER;

	return 0;
}

/*
	Fade out and stop streaming audio
*/ 

void FadeOutAudio()
{
	int v=0x10000000;
	if (!numSongs)
		return;

	while(v)
    {
		AUDIOMixer(v,0x30000000);
		v-=0x800000;
		_VidSync(1);
	}
	AUDIOStopStreamingAudio();
    AUDIOMixer(0x10000000,0x30000000);
}

/*
	Fade in and start streaming audio
*/ 

void FadeInAudio()
{
	int v=0;

	if (!numSongs)
		return;
	
    AUDIOMixer(0,0x30000000);
	AUDIOStartStreamingAudio();
	while(v!=0x10000000)
    {
		AUDIOMixer(v,0x30000000);
		v+=0x800000;
		_VidSync(1);
	}
}

/*
	Called by the game main loop! This handles the actual audio data loading
	on Samsung players,
*/

void HandleStreamingAudio()
{
	int sectLeft;
	static int err=0;
	static int n=0;

	switch(streamState)
	{
		case kWaitForRead:
			
			if (CDRequest&0x02)
			{
				kprintf("  Done\n");
				audSector+=n;
				streamCount-=n;
				streamAdr+=n*2048;
				if (audSector==audStartSector+audNumSectors)
					audSector=audStartSector;	
				if (!streamCount)
					streamState=kIdle;
				else
					streamState=kStartRead;
				return;	
			}

			if (!(CDRequest&0x01) || (CDRequest&0x80) || err<0)
			{
				kprintf("Retry (%04x/%d)\n",CDRequest,err);
				streamState=kStartRead;
			}
		break;
		
		case kStartRead:	
			sectLeft = audStartSector+audNumSectors-audSector;
			if (streamCount<=sectLeft)
				n=streamCount;
			else
				n=sectLeft;	
			kprintf(" read %d %d-%d\n",(int)(streamAdr-audiobuffer)/2048,audSector,audSector+n-1);
			err=CDReadSectors(audSector,n,streamAdr,0x80+isXA);
			streamState=kWaitForRead;
		break;
	}
}

#else

/*
	Empty function dummys for disabled streaming audio build!
*/

int InitStreamingAudio()
{
    AUDIOMixer(0x00000000,0x40000000);
	return 0;
}

int StartSong(int n)
{
}

void FadeInAudio()
{
}

void FadeOutAudio()
{
}

void HandleStreamingAudio()
{
}

#endif
